@@ -34,6 +34,7 @@ use crate::chain_error;
 use crate::chain_info;
 use crate::chain_warn;
 use crate::drop_result;
+use crate::executor::block_on_with_executor;
 use crate::executor::call_with_executor;
 use crate::filtered_io_error;
 use crate::fresh_error;
@@ -591,7 +592,7 @@ fn consume_directory<R: 'static + Future<Output = Result<()>>>(
                         };
 
                     match consumer(&name, contents) {
-                        Ok(None) => cleanup(Ok(()), &name, &to_consume),
+                        Ok(None) => cleanup(Ok(()), &name, to_consume),
                         Ok(Some(continuation)) => {
                             let name = name.to_owned();
                             let to_consume = to_consume.clone();
@@ -617,7 +618,7 @@ fn consume_directory<R: 'static + Future<Output = Result<()>>>(
                                 std::mem::drop(token);
                             });
                         }
-                        Err(e) => cleanup(Err(e), &name, &to_consume),
+                        Err(e) => cleanup(Err(e), &name, to_consume),
                     };
                 }
 
@@ -629,7 +630,6 @@ fn consume_directory<R: 'static + Future<Output = Result<()>>>(
     match std::fs::read_dir(&to_consume) {
         Ok(dirents) => {
             call_with_executor(|runtime| {
-                let _scope = runtime.enter();
                 let local_set = tokio::task::LocalSet::new();
                 consume_files(runtime, &local_set, dirents, &mut to_consume);
                 // Wait for any pending task before returning.
@@ -658,7 +658,7 @@ fn consume_directory<R: 'static + Future<Output = Result<()>>>(
 /// Creates `bucket` if it does not already exists.
 #[instrument(level = "debug", skip(bucket), err)]
 fn ensure_bucket_exists(bucket: &Bucket) -> Result<()> {
-    let bucket_location = bucket.location_blocking().map_err(|e| {
+    let bucket_location = block_on_with_executor(|| bucket.location()).map_err(|e| {
         chain_debug!(
             e,
             "failed to get buccket location",
@@ -670,16 +670,23 @@ fn ensure_bucket_exists(bucket: &Bucket) -> Result<()> {
         return Ok(());
     }
 
-    let result = if bucket.is_subdomain_style() {
-        Bucket::create_blocking
-    } else {
-        Bucket::create_with_path_style_blocking
-    }(
-        &bucket.name(),
-        bucket.region(),
-        bucket.credentials().clone(),
-        s3::bucket_ops::BucketConfiguration::private(),
-    );
+    let result = call_with_executor(|rt| {
+        if bucket.is_subdomain_style() {
+            rt.block_on(Bucket::create(
+                &bucket.name(),
+                bucket.region(),
+                bucket.credentials().clone(),
+                s3::bucket_ops::BucketConfiguration::private(),
+            ))
+        } else {
+            rt.block_on(Bucket::create_with_path_style(
+                &bucket.name(),
+                bucket.region(),
+                bucket.credentials().clone(),
+                s3::bucket_ops::BucketConfiguration::private(),
+            ))
+        }
+    });
 
     match result {
         Ok(response)
@@ -711,7 +718,7 @@ fn create_target(
     match target {
         S3(s3) => {
             let region = parse_s3_region_specification(&s3.region, s3.endpoint.as_deref());
-            let bucket_name = bucket_extractor(&s3);
+            let bucket_name = bucket_extractor(s3);
             let mut bucket = Bucket::new(bucket_name, region, creds)
                 .map_err(|e| chain_error!(e, "failed to create S3 bucket object", ?s3))?;
 
@@ -890,7 +897,7 @@ fn touch_blob(blob_name: &str, targets: &mut [Bucket]) -> Result<()> {
         // given that blob names are themselves percent encoded).
         let url_encoded_name = percent_encoding::utf8_percent_encode(
             &location_name,
-            &percent_encoding::NON_ALPHANUMERIC,
+            percent_encoding::NON_ALPHANUMERIC,
         );
 
         target.add_header(COPY_SOURCE, &url_encoded_name.to_string());
@@ -902,11 +909,9 @@ fn touch_blob(blob_name: &str, targets: &mut [Bucket]) -> Result<()> {
             match call_with_slow_logging(
                 Duration::from_secs(10),
                 || {
-                    target.put_object_with_content_type_blocking(
-                        &blob_name,
-                        &[],
-                        CHUNK_CONTENT_TYPE,
-                    )
+                    block_on_with_executor(|| {
+                        target.put_object_with_content_type(&blob_name, &[], CHUNK_CONTENT_TYPE)
+                    })
                 },
                 |duration| tracing::info!(?duration, ?blob_name, "slow S3 COPY"),
             ) {
@@ -1665,7 +1670,7 @@ impl CopierWorker {
         ]
         .iter()
         {
-            let is_empty = directory_is_empty_or_absent(&cleanup)
+            let is_empty = directory_is_empty_or_absent(cleanup)
                 .map_err(|e| chain_debug!(e, "failed to list directory", ?cleanup));
             if matches!(is_empty, Ok(true)) {
                 // It's not an error if this fails: we expect failures
@@ -1681,7 +1686,7 @@ impl CopierWorker {
             Duration::from_secs(60),
             || {
                 self.handle_ready_directory(
-                    &last_manifest_copy,
+                    last_manifest_copy,
                     &cache,
                     &targets,
                     creds.clone(),
@@ -1699,7 +1704,7 @@ impl CopierWorker {
             Duration::from_secs(60),
             || {
                 self.handle_staging_directory(
-                    &last_manifest_copy,
+                    last_manifest_copy,
                     state,
                     stale,
                     &cache,
@@ -1730,7 +1735,7 @@ impl CopierWorker {
             Duration::from_secs(60),
             || {
                 self.handle_ready_directory(
-                    &last_manifest_copy,
+                    last_manifest_copy,
                     &cache,
                     &targets,
                     creds,
@@ -1847,7 +1852,7 @@ impl CopierWorker {
         for fprint in shuffled {
             self.pace();
             touch_blob(
-                &replication_buffer::fingerprint_chunk_name(&fprint),
+                &replication_buffer::fingerprint_chunk_name(fprint),
                 &mut chunks_buckets,
             )?;
         }
@@ -2002,7 +2007,7 @@ impl CopierWorker {
 }
 
 /// Returns the path for the source db given the spool directory path.
-fn source_db_for_spool(spool_path: &Path) -> Result<PathBuf> {
+fn source_db_for_spool(spool_path: &Path) -> Result<Option<PathBuf>> {
     let file_name = spool_path
         .parent()
         .ok_or_else(|| fresh_warn!("Spool path does not have an inode component", ?spool_path))?
@@ -2011,7 +2016,7 @@ fn source_db_for_spool(spool_path: &Path) -> Result<PathBuf> {
         .to_str()
         .ok_or_else(|| fresh_warn!("Final spool path component is not valid utf-8", ?spool_path))?;
 
-    Ok(replication_buffer::restore_slashes(file_name)?.into())
+    Ok(replication_buffer::restore_slashes(file_name)?.map(|x| x.into()))
 }
 
 impl CopierSpoolState {
@@ -2254,7 +2259,7 @@ impl CopierBackend {
                     // Data in `ready_buffer` is only advisory, it's
                     // never incorrect to drop the work unit.
                     if let Some(state) = self.active_spool_paths.get(&ready) {
-                        self.queue_active_state(&state);
+                        self.queue_active_state(state);
                     }
                 },
                 // Errors only happen when there is no more sender.
@@ -2264,7 +2269,13 @@ impl CopierBackend {
             recv(self.maintenance) -> maintenance => match maintenance {
                 Ok(Join(spool_path, source_or)) => {
                     let mut source = match source_db_for_spool(&spool_path) {
-                        Ok(path) => path,
+                        Ok(Some(path)) => path,
+                        Ok(None) => if let Some(path) = &source_or {
+                            path.clone()
+                        } else {
+                            tracing::debug!(?spool_path, "unable to decode overlong source db path");
+                            return true;
+                        }
                         Err(err) => {
                             tracing::warn!(?spool_path, ?err, "failed to decode source db path");
                             return true;
