@@ -78,7 +78,7 @@ extern "C" {
 /// "@/path/to/config_file.json".
 #[tracing::instrument]
 pub fn parse_configuration_string(config: &str) -> Option<Options> {
-    if let Some(path) = config.strip_prefix("@") {
+    if let Some(path) = config.strip_prefix('@') {
         let contents = match std::fs::read(path) {
             Ok(contents) => contents,
             Err(e) => {
@@ -117,7 +117,7 @@ pub fn parse_configuration_string(config: &str) -> Option<Options> {
 /// an `Options` struct, or "@/path/to/config_file.json".
 #[tracing::instrument]
 pub fn load_configuration_from_env(var_name_or: Option<&str>) -> Option<Options> {
-    let var_name = var_name_or.unwrap_or(&VERNEUIL_CONFIG_ENV_VAR);
+    let var_name = var_name_or.unwrap_or(VERNEUIL_CONFIG_ENV_VAR);
 
     let os_value = std::env::var_os(var_name)?;
     let value = if let Some(value) = os_value.to_str() {
@@ -284,16 +284,13 @@ impl ReplicationProtoData {
     /// `Ok(None)` if the bytes encode protobuf data, but fail
     /// application-level validation.
     fn new(bytes: Vec<u8>) -> Result<Option<ReplicationProtoData>> {
-        use prost::Message;
-
-        let v1 = match manifest_schema::Manifest::decode(&*bytes)
-            .map_err(|e| chain_error!(e, "failed to decode proto bytes"))?
-            .v1
-        {
+        let v1 = match manifest_schema::Manifest::decode_and_validate(&*bytes, "bytes")?.v1 {
             Some(v1) => v1,
             None => return Ok(None),
         };
 
+        // The branches below are checked by `decode_and_validate`,
+        // but it doesn't hurt to be safe.
         let header_fprint = if let Some(fprint) = v1.header_fprint {
             fprint.into()
         } else {
