@@ -17,6 +17,7 @@ use crate::chain_error;
 use crate::chain_info;
 use crate::chain_warn;
 use crate::drop_result;
+use crate::executor::block_on_with_executor;
 use crate::fresh_error;
 use crate::manifest_schema::fingerprint_file_chunk;
 use crate::manifest_schema::hash_file_chunk;
@@ -251,13 +252,13 @@ impl Loader {
                 Credentials::default().map_err(|e| chain_error!(e, "failed to get credentials"))?;
 
             for source in remote {
-                if let Some(bucket) = create_source(&source, &creds, |s3| &s3.chunk_bucket)? {
+                if let Some(bucket) = create_source(source, &creds, |s3| &s3.chunk_bucket)? {
                     remote_sources.push(bucket);
                 }
             }
         }
 
-        cache_builder = apply_cache_replication_targets(cache_builder, &remote);
+        cache_builder = apply_cache_replication_targets(cache_builder, remote);
 
         // In tests, check every potential cache hits to make sure
         // they have the same value.
@@ -337,7 +338,7 @@ impl Loader {
             use std::io::ErrorKind;
 
             for source in &self.remote_sources {
-                if let Some(remote) = load_from_source(&source, &name)
+                if let Some(remote) = load_from_source(source, &name)
                     .map_err(|_| Error::new(ErrorKind::Other, "failed to fetch remote chunk"))?
                 {
                     let bytes = maybe_decompress(remote, fprint)
@@ -385,11 +386,11 @@ fn maybe_decompress(payload: Vec<u8>, fprint: Fingerprint) -> Result<Vec<u8>> {
     let hash_matches = |payload: &[u8]| -> bool {
         // Only check the first 64-bit half of the fingerprint, for
         // speed.
-        hash_file_chunk(&payload) == fprint.hash[0]
+        hash_file_chunk(payload) == fprint.hash[0]
     };
 
     let compare_hash = |payload: &[u8]| -> Result<()> {
-        let actual_hash = hash_file_chunk(&payload);
+        let actual_hash = hash_file_chunk(payload);
 
         if actual_hash == fprint.hash[0] {
             Ok(())
@@ -508,7 +509,7 @@ fn create_source(
     match source {
         S3(s3) => {
             let region = parse_s3_region_specification(&s3.region, s3.endpoint.as_deref());
-            let bucket_name = bucket_extractor(&s3);
+            let bucket_name = bucket_extractor(s3);
             let mut bucket = Bucket::new(bucket_name, region, creds.clone())
                 .map_err(|e| chain_error!(e, "failed to create chunks S3 bucket object", ?s3))?;
 
@@ -552,12 +553,14 @@ pub(crate) fn load_from_source(source: &Bucket, name: &str) -> Result<Option<Vec
     for i in 0..=LOAD_RETRY_LIMIT {
         match call_with_slow_logging(
             LOAD_REQUEST_TIMEOUT,
-            || source.get_object_blocking(name),
+            || block_on_with_executor(|| source.get_object(name)),
             |duration| tracing::info!(?duration, ?name, "slow S3 GET"),
         ) {
             Ok((payload, 200)) => return Ok(Some(payload)),
-            Ok((_, 404)) => return Ok(None),
-            Ok((body, code)) if code < 500 && code != 429 => {
+            // All callers propagate 404s as hard failures.  We might
+            // as try a little bit harder if we get a 404.
+            Ok((_, 404)) if i > 0 => return Ok(None),
+            Ok((body, code)) if code < 500 && ![404, 429].contains(&code) => {
                 return Err(
                     chain_error!((body, code), "failed to fetch chunk", %source.name, %name),
                 )
