@@ -162,7 +162,6 @@ use crate::filtered_io_error;
 use crate::fresh_error;
 use crate::fresh_info;
 use crate::instance_id;
-use crate::manifest_schema::fingerprint_v1_chunk_list;
 use crate::manifest_schema::Manifest;
 use crate::ofd_lock::OfdLock;
 use crate::process_id::process_id;
@@ -224,6 +223,15 @@ const SCRATCH_FILE_GRACE_PERIOD: Duration = Duration::from_secs(10);
 /// this grace period.
 const ORPHAN_TAP_FILE_GRACE_PERIOD: Duration = Duration::from_secs(24 * 3600);
 
+/// Contemporary file systems (unixen and windows) all let files and
+/// directories have up to 255 bytes or codepoints in their name.
+///
+/// Let invertibly mangled names go up to 230 *bytes*, just to avoid
+/// issues with boundary value.  In practice, invertible names are
+/// always strictly shorter than `MAX_MANGLED_NAME_LENGTH`, and
+/// uninvertible ones strictly longer.
+const MAX_MANGLED_NAME_LENGTH: usize = 230;
+
 /// Sets the default spooling directory for replication subdirectories,
 /// if it isn't already set.
 #[instrument]
@@ -442,19 +450,66 @@ pub(crate) fn reset_meta_copy_lock(spool_dir: PathBuf) -> Result<()> {
 fn replace_slashes(input: &str) -> String {
     const ESCAPED: percent_encoding::AsciiSet = percent_encoding::CONTROLS.add(b'#').add(b'%');
 
-    percent_encoding::utf8_percent_encode(&input, &ESCAPED)
+    percent_encoding::utf8_percent_encode(input, &ESCAPED)
         .map(|fragment| fragment.replace("/", "#"))
         .collect()
 }
 
-/// Reverses `replace_slashes`.
-pub(crate) fn restore_slashes(input: &str) -> Result<String> {
+/// Reverses `replace_slashes` if possible.
+pub(crate) fn restore_slashes(input: &str) -> Result<Option<String>> {
     let slashified = input.replace("#", "/");
 
-    Ok(percent_encoding::percent_decode_str(&slashified)
-        .decode_utf8()
-        .map_err(|e| chain_info!(e, "invalid utf-8 bytes"))?
-        .to_string())
+    // If the input string is more than `MAX_MANGLED_NAME_LENGTH`
+    // bytes, assume it had to be truncated.
+    if input.as_bytes().len() > MAX_MANGLED_NAME_LENGTH {
+        Ok(None)
+    } else {
+        Ok(Some(
+            percent_encoding::percent_decode_str(&slashified)
+                .decode_utf8()
+                .map_err(|e| chain_info!(e, "invalid utf-8 bytes"))?
+                .to_string(),
+        ))
+    }
+}
+
+fn mangle_string(string: &str) -> String {
+    let unslashified = replace_slashes(string);
+
+    // Avoid fencepost bugs and never generate a mangled string
+    // that's exactly MAX_MANGLED_NAME_LENGTH long.
+    if unslashified.as_bytes().len() < MAX_MANGLED_NAME_LENGTH {
+        return unslashified;
+    }
+
+    lazy_static::lazy_static! {
+        static ref MANGLE_PARAMS: umash::Params = umash::Params::derive(0, b"verneuil path mangling params");
+    }
+
+    // This fingerprint will be printed as 2x 16 hex char strings.
+    let fprint = MANGLE_PARAMS
+        .fingerprinter(0)
+        .write(unslashified.as_bytes())
+        .digest();
+
+    // MAX_MANGLED_NAME_LENGTH is a bit short of the expected
+    // limit on filenames; this lets us add a few extra bytes to
+    // make sure the mangled path exceeds MAX_MANGLED_NAME_LENGTH,
+    // even after round tripping to unicode codepoint and back to
+    // encoded bytes.
+    let fragment_len = 5 + (MAX_MANGLED_NAME_LENGTH - 32) / 2;
+    // The conversion back from truncated bytes to codepoint to utf-8
+    // bytes might add a couple bytes on each end, but
+    // MAX_MANGLED_NAME_LENGTH is comfortable less than 255, so the
+    // final mangled name is both definitely longer than
+    // MAX_MANGLED_NAME_LENGTH and shorter than 255.
+    let left = String::from_utf8_lossy(&unslashified.as_bytes()[..fragment_len]);
+    let right =
+        String::from_utf8_lossy(&unslashified.as_bytes()[unslashified.len() - fragment_len..]);
+    format!(
+        "{}{:016x}{:016x}{}",
+        left, fprint.hash[0], fprint.hash[1], right
+    )
 }
 
 /// Mangles a path to an extent database into a directory name:
@@ -467,8 +522,7 @@ fn mangle_path(path: &Path) -> Result<String> {
         .as_os_str()
         .to_str()
         .ok_or_else(|| fresh_error!("unable to convert canonical path to string", ?canonical))?;
-
-    Ok(replace_slashes(string))
+    Ok(mangle_string(string))
 }
 
 /// Constructs a unique human-readable filename for `fd`: we derive a
@@ -565,10 +619,10 @@ fn delete_dangling_replication_directories(mut parent: PathBuf) -> Result<()> {
         }
 
         let base_file = match restore_slashes(&name) {
-            Ok(base_file) => base_file,
+            Ok(Some(base_file)) => base_file,
             // If we couldn't decode the file name, it's probably
             // not ours?
-            Err(_) => continue,
+            Ok(None) | Err(_) => continue,
         };
 
         // If the file definitely does not exist, delete the
@@ -749,13 +803,14 @@ pub fn manifest_name_for_hostname_path(hostname_or: Option<&str>, path: &Path) -
         static ref PARAMS: umash::Params = umash::Params::derive(0, b"verneuil path params");
     }
 
+    #[allow(clippy::redundant_closure)] // we need to adjust hostname()'s lifetime
     let hostname: &str = hostname_or.unwrap_or_else(|| instance_id::hostname());
 
     let string = path
         .as_os_str()
         .to_str()
         .ok_or_else(|| fresh_error!("failed to convert path to string", ?path))?;
-    let path_hash = PARAMS.hasher(0).write(&string.as_bytes()).digest();
+    let path_hash = PARAMS.hasher(0).write(string.as_bytes()).digest();
 
     let name = format!(
         "{}-verneuil:{}:{:04x}/{}",
@@ -810,28 +865,13 @@ pub(crate) fn chunk_name_fingerprint(name: &str) -> Option<Fingerprint> {
 /// Returns Ok(None) if the file does not exist.
 #[instrument(level = "trace", err)]
 fn read_manifest_at_path(file_path: &Path) -> Result<Option<Manifest>> {
-    use prost::Message;
-
     let contents = match std::fs::read(file_path) {
         Ok(contents) => contents,
         Err(e) if e.kind() == ErrorKind::NotFound => return Ok(None),
         Err(e) => return Err(chain_error!(e, "failed to read manifest", ?file_path)),
     };
-    let manifest = Manifest::decode(&*contents)
-        .map_err(|e| chain_error!(e, "failed to parse proto manifest", ?file_path))?;
 
-    match &manifest.v1 {
-        Some(v1)
-            // We must have a v1 entry, the chunk fingerprints must match, and there
-            // must be an even number of u64 in the list (we need two per fingerprint).
-            if Some(fingerprint_v1_chunk_list(&v1.chunks).into()) == v1.contents_fprint
-                && (v1.chunks.len() % 2) == 0 =>
-        {
-            Ok(Some(manifest))
-        },
-        Some(_) => Err(fresh_error!("invalid chunk list", ?manifest)),
-        None => Err(fresh_error!("v1 format not found", ?manifest)),
-    }
+    Ok(Some(Manifest::decode_and_validate(&*contents, file_path)?))
 }
 
 /// Returns the path to the staging directory under `parent`; it may
@@ -1503,10 +1543,52 @@ fn replace_slashes_invertible() {
     // And it must be invertible.
     assert_eq!(
         HARD_STRING,
-        &restore_slashes(&converted).expect("must decode")
+        &restore_slashes(&converted)
+            .expect("must decode")
+            .expect("must be reversible")
     );
 }
 
+/// A short path should be invertible.
+#[test]
+fn mangle_path_short() {
+    let path = format!("/foo/bar/{}", String::from_utf8(vec![b'X'; 220]).unwrap());
+
+    let mangled = mangle_string(&path);
+
+    assert!(mangled.as_bytes().len() < MAX_MANGLED_NAME_LENGTH);
+    assert_eq!(
+        path,
+        restore_slashes(&mangled)
+            .expect("must decode")
+            .expect("must be reversible")
+    );
+}
+
+/// A long path should be clearly not invertible.
+#[test]
+fn mangle_path_long() {
+    let path = format!("/foo/bar/{}", String::from_utf8(vec![b'X'; 221]).unwrap());
+
+    let mangled = mangle_string(&path);
+
+    assert!(mangled.as_bytes().len() > MAX_MANGLED_NAME_LENGTH);
+    assert_eq!(None, restore_slashes(&mangled).expect("must decode"));
+}
+
+/// A long path should mangle to comfortable less than 255 bytes
+/// (NAME_MAX on most platforms).
+#[test]
+fn mangle_path_over_name_max() {
+    let path = format!("/foo/bar/{}", String::from_utf8(vec![b'X'; 1000]).unwrap());
+
+    let mangled = mangle_string(&path);
+
+    assert!(mangled.as_bytes().len() > MAX_MANGLED_NAME_LENGTH);
+    assert_eq!(240, mangled.as_bytes().len());
+    assert_eq!(None, restore_slashes(&mangled).expect("must decode"));
+}
+
 /// The mapping function should be constant across versions.
 #[test]
 fn test_manifest_name_for_hostname_path() {
